#ifndef OUTILS_SPECIFIQUES_H
#define OUTILS_SPECIFIQUES_H

#include "Etat.h"
#include "Predicat.h"
#include "Transition.h"
#include "Automate.h"
#include "PtrPartage.h"
#include <cassert>

//
// Etats specifiques au problme abord
//
class Pas
{
   int nb_;
public:
   Pas()
      : nb_(0)
   {
   }
   int nb_pas() const throw()
      { return nb_; }
   void marcher() throw()
      { ++nb_; }
};

class EtatMouvement
   : public Etat
{
   PtrPartage<Pas> pas_;
public:
   EtatMouvement(PtrPartage<Pas> pas)
      : pas_(pas)
   {
      assert(pas_);
   }
   int nb_pas() const throw()
      { return pas_->nb_pas(); }
   void dedans()
      { pas_->marcher(); }
};

template < >
   struct traits_dup<EtatMouvement>
   {
      typedef DupClonage categorie;
   };

struct VaAGauche
   : EtatMouvement
{
   VaAGauche(PtrPartage<Pas> pas)
      : EtatMouvement(pas)
   {
   }
   VaAGauche* cloner() const
      { return new VaAGauche(*this); }
   void entrer();
   void dedans();
   void sortir();
};

struct VaADroite
   : EtatMouvement
{
   VaADroite(PtrPartage<Pas> pas)
      : EtatMouvement(pas)
   {
   }
   VaADroite* cloner() const
      { return new VaADroite(*this); }
   void entrer();
   void dedans();
   void sortir();
};

//
// Prdicats spcifiques
//
struct ChangementDirection
   : Predicat
{
   bool verifier(PtrPartage<Etat> p) const
      { return partageable_cast<EtatMouvement>(p)->nb_pas() % 5 == 0; }
};

struct Essouflement
   : Predicat
{
   bool verifier(PtrPartage<Etat> p) const
      { return partageable_cast<EtatMouvement>(p)->nb_pas() >= 30; }
};

class AutomateAllerRetour
   : public Automate
{
   void construire();
};

#endif
