#ifndef PTR_PARTAGE_H
#define PTR_PARTAGE_H

#include "EstConvertible.h"
#include "Clonable.h"
#include "Testable.h"
#include <utility> // std::swap()

template <class T>
   class PtrPartage
      : public Testable
   {
      class Partageable
      {
         unsigned long cptref_;
      public:
         Partageable()
            : cptref_(1UL)
         {
         }
         void AddRef()
            { ++cptref_; }
         void Release()
            { if (--cptref_ == 0) delete this; }
      protected:
         virtual ~Partageable() throw()
            { }
      };
      class DecorateurPartageable
         : public Partageable
      {
         T *ptr_;
      public:
         DecorateurPartageable(T *p = 0) throw()
            : ptr_(p) { }
         T *get() const throw()
            { return ptr_; }
         operator bool() const throw()
            { return get() != 0; }
      };
      DecorateurPartageable *ptr_;
      class AutoRelacheur
      {
         DecorateurPartageable *tmpptr_;
      public:
         AutoRelacheur(DecorateurPartageable *p)
            : tmpptr_(p)
         {
            if (tmpptr_) tmpptr_->AddRef();
         }
         ~AutoRelacheur() throw()
            { if (tmpptr_) tmpptr_->Release(); }
      };
   public:
      explicit PtrPartage(T *p = 0)
         : ptr_(0)
      {
         ptr_ = new DecorateurPartageable(p);
      }
      PtrPartage(const PtrPartage &p)
         : ptr_(p.ptr_)
      {
         if (ptr_) ptr_->AddRef();
      }
      T* get() const throw()
         { return ptr_? ptr_->get() : 0; }
      //
      // Ici, on est forcs de duplique le point. Prudence
      //
      template <class U>
         PtrPartage(const PtrPartage<U> &p)
            : ptr_(0)
         {
            if (p.get())
            {
               ptr_ = new DecorateurPartageable(
                  dupliquer(
                     dynamic_cast<T*>(p.get())
                  )
               );
            }
         }
      void swap (PtrPartage &p)
         { std::swap (ptr_, p.ptr_); }
      PtrPartage& operator=(const PtrPartage &p)
      {
         PtrPartage(p).swap(*this);
         return *this;
      }
      template <class U>
         PtrPartage& operator=(const PtrPartage<U> &p)
         {
            PtrPartage (p).swap(*this);
            return *this;
         }
      ~PtrPartage() throw()
      {
         if (ptr_) ptr_->Release();
      }
      T& operator*()
         { return *(ptr_->get()); }
      const T& operator*() const // indirection importante si T est abstrait
         { return *(ptr_->get()); }
      T* operator->()
      {
         if (!ptr_) return 0;
         AutoRelacheur ar(ptr_);
         return ptr_->get();
      }
      const T* operator->() const
      {
         if (!ptr_) return 0;
         AutoRelacheur ar(ptr_);
         return ptr_->get();
      }
      //
      // RAFFINER
      //
   private:
      bool_type validite() const
      {
         return get()?
            true_value() : false_value();
      }
   };

template <class T, class U>
   PtrPartage<T> partageable_cast(const PtrPartage<U> &p)
   {
      return PtrPartage<T>(p);
   }

#endif

